# -*- python -*-
# $RCSfile: boundarycondition.py,v $
# $Revision: 1.78 $
# $Author: rlua $
# $Date: 2007-09-28 03:20:51 $


# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

from ooflib.SWIG.common import switchboard
from ooflib.SWIG.common import ooferror
from ooflib.common import debug
from ooflib.common import labeltree
from ooflib.common import utils
from ooflib.common.IO import automatic
from ooflib.common.IO import oofmenu
from ooflib.common.IO import parameter
from ooflib.common.IO import reporter
from ooflib.common.IO import whoville
from ooflib.common.IO import xmlmenudump
from ooflib.engine import bdycondition
from ooflib.engine import skeletoncontext
from ooflib.engine.IO import meshmenu
from ooflib.engine.IO import meshparameters
from ooflib.common import parallel_enable
if parallel_enable.enabled():
    from ooflib.engine.IO import boundaryconditionIPC
import ooflib.engine.mesh
import ooflib.engine.profile
import types

bcmenu = oofmenu.OOFMenuItem(
    'Boundary_Conditions',
    cli_only=1,
    help="Create and manipulate boundary conditions on Meshes.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu/bc.xml')
    )

meshmenu.meshmenu.addItem(bcmenu)


def _buildbc(menuitem, name, mesh, condition):
    msg = condition.check(mesh)
    if msg:
        raise ooferror.ErrSetupError(msg)
    else:
        if parallel_enable.enabled():
            boundaryconditionIPC.ipcbcmenu.buildbc(name=name,mesh=mesh,condition=condition)
        else:
            condition.add_to_mesh(name, mesh)
            switchboard.notify("mesh changed", ooflib.engine.mesh.meshes[mesh])

def bcnameResolver(param, name):
    if param.automatic():
        basename = 'bc'
    else:
        basename = name
    meshname = param.group['mesh'].value
    if meshname is not None:
        meshpath = labeltree.makePath(meshname)
        meshcontext = ooflib.engine.mesh.meshes[meshpath]
        return meshcontext.uniqueBCName(basename)

bcmenu.addItem(oofmenu.OOFMenuItem(
    'New',
    callback=_buildbc,
    params=parameter.ParameterGroup(
    parameter.AutomaticNameParameter('name',
                                     value=automatic.automatic,
                                     tip="Name of the new boundary condition",
                                     resolver=bcnameResolver),
    whoville.WhoParameter('mesh', ooflib.engine.mesh.meshes,
                          tip='Mesh to which to apply the boundary condition.'),
    parameter.RegisteredParameter("condition", bdycondition.BC,
                                  tip="The new boundary condition.")),
    help='Create a new boundary condition.',
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu/new_bc.xml')
    ))


def _bcrename(menuitem, mesh, bc, name):
    if parallel_enable.enabled():
        boundaryconditionIPC.ipcbcmenu.Rename(mesh=mesh,bc=bc,name=name)
    else:
        meshcontext = ooflib.engine.mesh.meshes[mesh]
        meshcontext.renameBdyCondition(bc, name)

bcmenu.addItem(oofmenu.OOFMenuItem(
    "Rename",
    callback=_bcrename,
    params=parameter.ParameterGroup(
    whoville.WhoParameter("mesh", ooflib.engine.mesh.meshes,
                          tip=parameter.emptyTipString),
    parameter.StringParameter('bc', tip='Old name of the boundary condition'),
    parameter.StringParameter('name', tip='New name of the boundary condition')
##     parameter.AutomaticNameParameter('name', value=automatic.automatic,
##                                      tip="New name for the bc",
##                                      resolver=bdycondition.bcnameResolver)
    ),
    help="Give a boundary condition a new name.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu/rename_bc.xml')
    ))

# Remove a boundary condition from the given mesh.
def _bcremove(self, mesh, name):
    if parallel_enable.enabled():
        boundaryconditionIPC.ipcbcmenu.Delete(mesh=mesh,name=name)
    else:
        meshcontext = ooflib.engine.mesh.meshes[mesh]
        meshcontext.rmBdyCondition(name)
        switchboard.notify("mesh changed", ooflib.engine.mesh.meshes[mesh])

bcmenu.addItem(oofmenu.OOFMenuItem(
    "Delete",
    callback=_bcremove,
    params = [whoville.WhoParameter('mesh', ooflib.engine.mesh.meshes,
                                    tip="Name of the mesh."),
              parameter.StringParameter('name', tip="Name of the boundary condition to be deleted.")],
    help="Remove this condition from the mesh.",
    discussion="<para>Remove this boundary condition from the &mesh;.</para>"
    ))

# Toggle whether BC is enabled
# TODO: need IPC version?
##def _bcenable(self, mesh, name):
##    meshcontext = oof2.engine.mesh.meshes[mesh]
##    bndycond = meshcontext.getBdyCondition(name)
##    bndycond.toggle_explicit_disable()

##bcmenu.addItem(oofmenu.OOFMenuItem(
##    "EnableDisable",
##    callback = _bcenable,
##    params = [whoville.WhoParameter('mesh', oof2.engine.mesh.meshes,
##                                    tip="Name of the mesh."),
##               parameter.StringParameter('name', tip="Boundary condition being enabled or disabled")],
##    help="Enable or disable a boundary condition.",
##    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu/enabledisable_bc.xml')
##    ))

def _bcenable(self, mesh, name):
    if parallel_enable.enabled():
        boundaryconditionIPC.ipcbcmenu.Enable(mesh=mesh,name=name)
        return
    meshcontext = ooflib.engine.mesh.meshes[mesh]
    bndycond = meshcontext.getBdyCondition(name)
    bndycond.explicitly_enable()

def _bcdisable(self, mesh, name):
    if parallel_enable.enabled():
        boundaryconditionIPC.ipcbcmenu.Disable(mesh=mesh,name=name)
        return
    meshcontext = ooflib.engine.mesh.meshes[mesh]
    bndycond = meshcontext.getBdyCondition(name)
    bndycond.explicitly_disable()

bcmenu.addItem(oofmenu.OOFMenuItem(
    "Enable",
    callback=_bcenable,
    params=[whoville.WhoParameter('mesh', ooflib.engine.mesh.meshes,
                                  tip="Name of the mesh."),
            parameter.StringParameter('name', tip="Boundary condition being enabled.")],
    help="Enable an explicitly disabled boundary condition.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu/enable_bc.xml')
    ))

bcmenu.addItem(oofmenu.OOFMenuItem(
    "Disable",
    callback=_bcdisable,
    params=[whoville.WhoParameter('mesh', ooflib.engine.mesh.meshes,
                                  tip="Name of the mesh."),
            parameter.StringParameter('name', tip="Boundary condition being disabled.")],
    help="Disable a boundary condition.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu/disable_bc.xml')
    ))

# Copy a boundary condition to a new boundary (for possibly another mesh).
def _bccopy(self, current, mesh, bc, name, boundary):
    if parallel_enable.enabled():
        boundaryconditionIPC.ipcbcmenu.Copy(current=current,mesh=mesh,bc=bc,name=name,boundary=boundary)
    else:
        meshcontext = ooflib.engine.mesh.meshes[current]
        bndycond = meshcontext.getBdyCondition(bc)
        newbc = bndycond.copy(boundary)
        msg = newbc.check(mesh)
        if msg:
            raise ooferror.ErrSetupError(msg)
        else:
            newbc.add_to_mesh(name, mesh)
            switchboard.notify("mesh changed", ooflib.engine.mesh.meshes[mesh])
    
bcmenu.addItem(oofmenu.OOFMenuItem(
    "Copy",
    callback=_bccopy,
    params=parameter.ParameterGroup(
    whoville.WhoParameter("current", ooflib.engine.mesh.meshes,
                          tip="Copy the boundary condition from this mesh."),
    whoville.WhoParameter("mesh", ooflib.engine.mesh.meshes,
                          tip="Copy the boundary condition to this mesh."),
    parameter.StringParameter("bc", tip="Boundary condition being copied"),
    parameter.AutomaticNameParameter("name", value=automatic.automatic,
                                     resolver=bcnameResolver,
                                     tip="Name of the new boundary condition."),
    meshparameters.MeshBoundaryParameter("boundary",
                                         tip="Boundary to which the copied BC will apply.")
    ),
    help="Copy a boundary condition from one mesh to another.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu/copy_bc.xml')
    ))

# Copy all the boundary conditions from one mesh to another, copying
# them to the same boundaries, of course.
def _bccopyall(self, current, mesh):
    if parallel_enable.enabled():
        boundaryconditionIPC.ipcbcmenu.Copy_All(current=current,mesh=mesh)
    else:
        meshcontext = ooflib.engine.mesh.meshes[current]
        allbcs = meshcontext.allBoundaryConds()
        for (name, cond) in allbcs:
            #Skip the auxilliary FloatBCs for the periodic boundary conditions.
            #These are created when PeriodicBC.add_to_mesh is called.
            #TODO: May need a better check later. See PeriodicBC.add_to_mesh
            #Interface branch: Do the same for the invisible Float BCs
            #associated with interfaces.
            if name.find("aux_pointbdy")==0 or name.find("_cntnty_")==0:
                continue
            msg = cond.check(mesh)
            if msg:
                raise ooferror.ErrSetupError(msg)
                return
        for (name, cond) in allbcs:
            #TODO: May need a better check later.
            #Interface branch, update.
            if name.find("aux_pointbdy")==0 or name.find("_cntnty_")==0:
                continue
            newbc = cond.copy(cond.boundary)
            newbc.add_to_mesh(name, mesh)
        switchboard.notify("mesh changed", ooflib.engine.mesh.meshes[mesh])

bcmenu.addItem(oofmenu.OOFMenuItem(
    "Copy_All",
    callback=_bccopyall,
    params=[
    whoville.WhoParameter("current", ooflib.engine.mesh.meshes,
                          tip="Copy boundary conditions from this mesh."),
    whoville.WhoParameter("mesh", ooflib.engine.mesh.meshes,
                          tip="Target mesh for the boundary conditions.")],
    help="Copy all the boundary conditions from one mesh to another.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu/copyall_bc.xml')))


def _bcedit(menuitem, name, mesh, condition):
    if parallel_enable.enabled():
        boundaryconditionIPC.ipcbcmenu.Edit(name=name,mesh=mesh,condition=condition)
    else:
        meshctxt = ooflib.engine.mesh.meshes[mesh]
        oldcond = meshctxt.getBdyCondition(name)
        msg = condition.check(mesh, exclude=oldcond)
        if msg:
            raise ooferror.ErrSetupError(msg)
        else:
            meshctxt.rmBdyCondition(name)
            condition.add_to_mesh(name, mesh)
            switchboard.notify("mesh changed", ooflib.engine.mesh.meshes[mesh])
    
bcmenu.addItem(oofmenu.OOFMenuItem(
    "Edit",
    callback=_bcedit,
    params = [parameter.StringParameter("name",
                                        tip="Name of the boundary condition."),
              # No resolver, name must exist
              whoville.WhoParameter("mesh", ooflib.engine.mesh.meshes,
                                    tip=parameter.emptyTipString),
              parameter.RegisteredParameter("condition", bdycondition.BC,
                                            tip=parameter.emptyTipString)],
    help="Change the attributes of a boundary condition.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu/edit_bc.xml')))

                 
    


##############################
### Similar for Profiles.
##############################

pfmenu = oofmenu.OOFMenuItem(
    'Profiles',
    cli_only=1,
    help="Create and manipulate boundary condition profiles.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu/profile.xml'))

meshmenu.meshmenu.addItem(pfmenu)

def pfnameResolver(param, name):
    if param.automatic():
        basename = 'profile'
    else:
        basename = name
    return ooflib.engine.profile.AllProfiles.uniqueName(basename)

def newNamedProfile(menuitem, name, profile):
    if parallel_enable.enabled():
        boundaryconditionIPC.ipcpfmenu.New(name=name,profile=profile)
    else:
        profile.set_name(name)

pfmenu.addItem(oofmenu.OOFMenuItem(
    'New',
    callback=newNamedProfile,
    params=[
    parameter.AutomaticNameParameter('name',
                                     value=automatic.automatic,
                                     tip="Name for the new Profile.",
                                     resolver=pfnameResolver),
    parameter.RegisteredParameter('profile', ooflib.engine.profile.CoreProfile,
                                  tip='The new Profile.')
    ],
    help='Construct a new named Profile.',
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu/newprofile.xml')
                                   ))


# Remove a profile from the profile manager.
def _pfremove(self, name):
    if parallel_enable.enabled():
        boundaryconditionIPC.ipcpfmenu.Delete(name=name)
    else:
        pf = ooflib.engine.profile.AllProfiles[name]
        if pf.deletable():
            del ooflib.engine.profile.AllProfiles[name]
        else:
            reporter.warn("Profile '%s' cannot be deleted because it is still being used." % name)

pfmenu.addItem(oofmenu.OOFMenuItem(
    "Delete",
    help="Delete a profile from the system.",
    callback=_pfremove,
    params = [parameter.StringParameter('name', tip="Name of the profile to be deleted.")],
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu/deleteprofile.xml')
    ))


def _pfrename(self, profile, name):
    if parallel_enable.enabled():
        boundaryconditionIPC.ipcpfmenu.Rename(profile=profile,name=name)
    else:
        ooflib.engine.profile.AllProfiles.rename(profile, name)

pfmenu.addItem(oofmenu.OOFMenuItem(
    "Rename",
    help="Rename a profile.",
    callback=_pfrename,
    params = [
    parameter.StringParameter('profile', tip="Old name of the profile."),
    parameter.StringParameter('name', tip='New name of the profile')],
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu/renameprofile.xml')
    ))


def _pfcopy(self, profile, name):
    if parallel_enable.enabled():
        boundaryconditionIPC.ipcpfmenu.Copy(profile=profile,name=name)
    else:
        old_profile = ooflib.engine.profile.AllProfiles[profile]
        new_profile = old_profile.clone()
        new_profile.set_name(name)

pfmenu.addItem(oofmenu.OOFMenuItem(
    "Copy",
    help="Make a copy of a profile.",
    callback=_pfcopy,
    params = [
    parameter.StringParameter('profile', tip="Name of the profile to be copied."),
    parameter.AutomaticNameParameter('name', resolver=pfnameResolver,
                                     tip="Name for the new profile.") ],
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu/copyprofile.xml')
    ))


def _pfedit(self, name, profile):
    if parallel_enable.enabled():
        boundaryconditionIPC.ipcpfmenu.Edit(name=name,profile=profile)
    else:
        ooflib.engine.profile.AllProfiles.replaceProfile(name, profile)

pfmenu.addItem(oofmenu.OOFMenuItem(
    "Edit",
    help="Replace a profile with a new profile.",
    callback=_pfedit,
    params = [
    parameter.StringParameter("name", tip="Name of the profile to be replaced."),
    # No resolver, name must exist
    parameter.RegisteredParameter("profile", ooflib.engine.profile.CoreProfile,
                                  tip="The replacement profile.")],
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu/editprofile.xml')
    ))

