# -*- python -*-
# $RCSfile: meshmenu.py,v $
# $Revision: 1.147 $
# $Author: langer $
# $Date: 2008-01-16 17:07:30 $


# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

from ooflib.SWIG.common import config
from ooflib.SWIG.common import ooferror
from ooflib.SWIG.common import switchboard
from ooflib.SWIG.common.IO import progressbar
from ooflib.SWIG.engine import masterelement
from ooflib.common import debug
from ooflib.common import enum
from ooflib.common import labeltree
from ooflib.common import microstructure
from ooflib.common import parallel_enable
from ooflib.common import quit
from ooflib.common import utils
from ooflib.common.IO import automatic
from ooflib.common.IO import datafile
from ooflib.common.IO import mainmenu
from ooflib.common.IO import microstructureIO
from ooflib.common.IO import oofmenu
from ooflib.common.IO import parameter
from ooflib.common.IO import reporter
from ooflib.common.IO import whoville
from ooflib.common.IO import xmlmenudump
from ooflib.engine import adaptivemeshrefinement
from ooflib.engine import fieldinit
from ooflib.engine import meshcrosssection
from ooflib.engine import meshmod
from ooflib.engine import skeletoncontext
from ooflib.engine import skeletonelement
from ooflib.engine.IO import meshIO
from ooflib.engine.IO import meshparameters
from ooflib.engine.IO import skeletonIO
#from ooflib.engine.IO import subproblemmenu #Not referenced in this file(?)


if parallel_enable.enabled():
    from ooflib.engine.IO import meshIPC
import ooflib.engine.mesh
import types
import string

OOF = mainmenu.OOF
meshmenu = mainmenu.OOF.addItem(oofmenu.OOFMenuItem(
    'Mesh',
    cli_only=1,
    help='Tools for creating and manipulating Meshes.',
    discussion="""<para>
    The <command>Mesh</command> menu contains tools for creating and
    manipulating finite element &meshes;, including methods for
    defining &fields; and determining which &equations; to <link
    linkend='MenuItem:OOF.Solver.Solve'>solve</link>.
    </para>"""))

####################

# Look for an enclosing mesh parameter -- if not found, use the
# enclosing skeleton parameter.  Mesh copying needs the first case,
# new mesh construction needs the second.

def meshNameResolver(param, startname):
    if param.automatic():
        basename = 'mesh'
    else:
        basename = startname

    try:
        meshname = param.group['mesh'].value
    except IndexError:
        skelname = param.group['skeleton'].value
        skelpath = labeltree.makePath(skelname)
    else:
        skelpath = labeltree.makePath(meshname)[:-1]

    return ooflib.engine.mesh.meshes.uniqueName(skelpath + [basename])

###################################

def newMesh(menuitem, name, skeleton, element_types):
    if parallel_enable.enabled():
        # skeleton is a string!
        # There is an issue with progress bar.
        # The following ASSUMES there are exactly three element_types:
        #(D_typename, T_typename and Q_typename, for edgement, Tri and Quad)
        meshIPC.parallel_newMesh(name,skeleton,
                                 element_types[0].name,
                                 element_types[1].name,
                                 element_types[2].name)
    else:
        ## TODO CLEANUP: Why are there so many separate progress bar
        ## function calls?  Why are some of them reporter functions
        ## and some progressbar functions?
        pbar = menuitem.getProgressBar("continuous")
        pbar.set_message("Preparing to create new mesh...")
        reporter.start_progressbar()
        try:
            edict = {}
            for eltype in element_types:
                el = masterelement.getMasterElementFromEnum(eltype)
                edict[el.nsides()] = el
            skelpath = labeltree.makePath(skeleton)
            skel = skeletoncontext.skeletonContexts[skelpath].getObject()
            femesh = skel.femesh(edict,
                                 skeletonelement.SkeletonElement.realmaterial,
                                 skeleton)
            if femesh is not None:
                meshctxt = ooflib.engine.mesh.meshes.add(
                    skelpath+[name], femesh,
                    parent=skeletoncontext.skeletonContexts[skelpath],
                    skeleton=skel,
                    elementdict=edict,
                    materialfactory=skeletonelement.SkeletonElement.realmaterial)
                meshctxt.createDefaultSubProblem()
                    
            pbar.set_progress(1)
            if pbar.query_stop()!=0 or pbar.get_success() is progressbar.FAILED:
                pbar.set_failure()
                pbar.set_message("Failed")
            else:
                pbar.set_success()
                pbar.set_message("Succeeded")
        finally:
            reporter.progressbar_completed()
            reporter.end_progressbar()

class MasterElementTypesParameter(enum.ListOfEnumsParameter):
    def __init__(self, name, value=None, default=None, tip=None):
        enum.ListOfEnumsParameter.__init__(
            self, name, masterelement.getMasterElementEnumClasses(),
            value, default, tip)
    def clone(self):
        return self.__class__(self.name, self.value, self.default, self.tip)
    def valueDesc(self):
        return "A list of element types."
    

newmeshcmd = meshmenu.addItem(oofmenu.OOFMenuItem(
    'New',
    callback=newMesh,
    threadable=oofmenu.THREADABLE,
    params=parameter.ParameterGroup(
    whoville.AutoWhoNameParameter('name', value=automatic.automatic,
                                  resolver=meshNameResolver,
                                  tip="Name of the new Mesh"),
    whoville.WhoParameter('skeleton', skeletoncontext.skeletonContexts,
                          tip=parameter.emptyTipString),
    MasterElementTypesParameter('element_types',
                                tip='A list of finite element types'),
##    parameter.BooleanParameter('split_interface', value=0,
##                               tip='Split the mesh along interfaces?')
    ),
    help='Create a new Mesh from a Skeleton.',
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu/newmesh.xml')
    ))

# The element_types parameter in the New Mesh menu item needs to be
# recreated whenever new MasterElement types are defined.
    
def buildNewMeshCmd():
    params = parameter.ParameterGroup(
        newmeshcmd.get_arg('name'),
        newmeshcmd.get_arg('skeleton'),
        MasterElementTypesParameter('element_types'))
    newmeshcmd.replace_args(params)

switchboard.requestCallback("new master element", buildNewMeshCmd)

#####################################

def renameMesh(menuitem, mesh, name):
    if parallel_enable.enabled():
        meshIPC.ipcmeshmenu.Rename(mesh=mesh,name=name)
        return

    oldmeshpath = labeltree.makePath(mesh)
    themesh = ooflib.engine.mesh.meshes[oldmeshpath]
    themesh.reserve()
    themesh.begin_writing()
    try:
        themesh.rename(name, exclude=oldmeshpath[-1])
    finally:
        themesh.end_writing()
        themesh.cancel_reservation()

meshmenu.addItem(oofmenu.OOFMenuItem(
    'Rename',
    callback=renameMesh,
    threadable=oofmenu.THREADABLE,
    params=[whoville.WhoParameter('mesh', ooflib.engine.mesh.meshes,
                                  tip=parameter.emptyTipString),
            whoville.WhoNameParameter('name', value='',
                                      tip='New name for the mesh.')
            ],
    help="Rename a Mesh.",
    discussion="<para> Assign a new name to a &mesh;. </para>"))

#######################################

def deleteMesh(menuitem, mesh):
    if parallel_enable.enabled():
        meshIPC.ipcmeshmenu.Delete(mesh=mesh)
        return

    meshctxt = ooflib.engine.mesh.meshes[mesh]

    subproblems = meshctxt.subproblems()
    for subproblem in subproblems:
        subproblem.begin_writing()
        try:
            subproblem.destroy()
        finally:
            subproblem.end_writing()
    
    meshctxt.reserve()
    meshctxt.begin_writing()
    try:
        meshctxt.destroy()         # removes mesh from ooflib.engine.mesh.meshes
    finally:
        meshctxt.end_writing()
        meshctxt.cancel_reservation()

meshmenu.addItem(oofmenu.OOFMenuItem(
    'Delete',
    threadable=oofmenu.THREADABLE,
    callback=deleteMesh,
    params=[whoville.WhoParameter('mesh', ooflib.engine.mesh.meshes,
                                  tip=parameter.emptyTipString)
    ],
    help="Delete a Mesh.",
    discussion="""<para>
    Delete a &mesh;.  Its &skel; and &micro; are
    <emphasis>not</emphasis> deleted.
    </para>"""
    ))

#######################################

def copyMesh(menuitem, mesh, name,
             copy_field, copy_equation, copy_bc):
    if parallel_enable.enabled():
        meshIPC.ipcmeshmenu.Copy(mesh=mesh,name=name,
                                 copy_field=copy_field, copy_equation=copy_equation, copy_bc=copy_bc)
        return

    notifications = []
    basemesh = ooflib.engine.mesh.meshes[mesh]
    basemesh.begin_reading()
    try:
        edict = basemesh.elementdict
        copiedmeshname = name

        skel = basemesh.getSkeleton()
        skelpath = labeltree.makePath(basemesh.path())[:-1]

        #Interface branch, pass skeleton path to femesh
        copiedfemesh = skel.femesh(edict, basemesh.materialfactory,
                                   skelpath[0]+":"+skelpath[1])
        newmesh = ooflib.engine.mesh.meshes.add(
            skelpath+[copiedmeshname],
            copiedfemesh,
            parent=skeletoncontext.skeletonContexts[skelpath],
            skeleton=skel, elementdict=edict,
            materialfactory=basemesh.materialfactory)
        newmesh.reserve()
        newmesh.begin_writing()
        try:
            copiedmesh = skelpath+[copiedmeshname]
            copiedmeshfullname = string.join(copiedmesh,":")
            for subpctxt in basemesh.subproblems():
                subp = subpctxt.getObject()
                newsubp = subpctxt.getObject().clone() # CSubProblem object
                newsubpctxt = newmesh.newSubProblem(
                    newsubp, copiedmesh+[subpctxt.name()])
                newsubpfullname = newsubpctxt.path()
                if copy_field:
                    for field in subpctxt.all_compound_fields():
                        newsubp.define_field(field)
                        notifications.append(
                            ("field defined", newsubpfullname, field.name(), 1))
                        newsubp.acquire_field_data(field, subp)
                        if subp.is_active_field(field):
                            newsubp.activate_field(field)
                            notifications.append(
                                ("field activated", newsubpfullname,
                                 field.name(), 1))
                if copy_equation:
                    for eqn in subpctxt.all_equations():
                        newsubp.activate_equation(eqn)
                        notifications.append(
                            ("equation activated",
                             newsubpfullname, eqn.name(), 1))
            # end loop over subproblems
            if copy_field:
                for field in newmesh.all_subproblem_fields():
                    if basemesh.femesh().in_plane(field):
                        newmesh.set_in_plane_field(field, 1)
                        notifications.append(("field inplane",
                                              copiedmeshfullname, field.name(),
                                              1))
            if copy_bc:
                for (bcname, bc) in basemesh.allBoundaryConds():
                    #Interface branch
                    #Don't copy the invisible Float BCs associated with
                    #interfaces. (see femesh.spy)
                    if bcname.find('_cntnty_')==0:
                        continue
                    copied = bc.copy(bc.boundary)
                    copied.add_to_mesh(bcname, copiedmesh)
        finally:
            newmesh.end_writing()
            newmesh.cancel_reservation()
    finally:
        basemesh.end_reading()

    for n in notifications:
        switchboard.notify(*n)


meshmenu.addItem(oofmenu.OOFMenuItem(
    'Copy', callback=copyMesh,
    threadable=oofmenu.THREADABLE,
    params= parameter.ParameterGroup(
    whoville.WhoParameter('mesh', ooflib.engine.mesh.meshes,
                          tip=parameter.emptyTipString),
    whoville.AutoWhoNameParameter('name', value=automatic.automatic,
                                  resolver=meshNameResolver,
                                  tip="Name of the copied Mesh. Use automatic selection, or type in a name."),
    parameter.BooleanParameter('copy_field', value=1, tip='Copy fields?'),
    parameter.BooleanParameter('copy_equation', value=1, tip='Copy equation?'),
    parameter.BooleanParameter('copy_bc', value=1,
                               tip='Copy boundary conditions?') ),
    help="Copy a Mesh.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu/copymesh.xml')
    ))

#######################################


# Copy the field state (definitions, active-ness, planarity) of one
# mesh into another.  This is the backwards-compatible deprecated
# version that uses the Mesh's default subproblem.  The preferred
# version is in subproblemmenu.py.

def _copyFieldState(menuitem, source, target):
    if source == target:
        raise ooferror.ErrUserError('Source and target must differ!')
    if parallel_enable.enabled():
        meshIPC.ipcmeshmenu.Copy_Field_State(source=source,target=target)
        return

    notifications = []
    source_mesh = ooflib.engine.mesh.meshes[source]
    target_mesh = ooflib.engine.mesh.meshes[target]
    source_subp = source_mesh.get_default_subproblem()
    target_subp = target_mesh.get_default_subproblem()
    source_subp.begin_reading()
    target_subp.reserve()
    target_subp.begin_writing()
    try:
        source_obj = source_subp.getObject()
        target_obj = target_subp.getObject()
        source_fields = source_subp.all_compound_fields()
        target_fields = target_subp.all_compound_fields()

     

        # Undefine all the fields in the target that are not in the source.
        for f in target_fields:
            if not source_obj.is_defined_field(f):
                target_obj.undefine_field(f)
                notifications.append(
                    ("field defined", target_subp.path(), f.name(), 0))

        for f in source_fields:
            # Definition.
            if not target_obj.is_defined_field(f):
                target_obj.define_field(f)
                notifications.append(
                    ("field defined", target_subp.path(), f.name(), 1))

            # Activation.
            if source_obj.is_active_field(f):
                if not target_obj.is_active_field(f):
                    target_obj.activate_field(f)
                    notifications.append(
                        ("field activated", target_subp.path(), f.name(), 1))
            else:
                if target_obj.is_active_field(f):
                    target_obj.deactivate_field(f)
                    notifications.append(
                        ("field activated", target_subp.path(), f.name(), 0))

            # Planarity.
            inplane = source_mesh.femesh().in_plane(f)
            if target_mesh.femesh().in_plane(f) != inplane:
                target_mesh.set_in_plane_field(f, inplane)
                notifications.append(("field inplane", target, f.name(),
                                      inplane))
            try:
                initializer = source_mesh.initializers[f]
            except KeyError:
                pass
            else:
                target_mesh.initialize_field(f, initializer)
                notifications.append(("field initialized"))
    finally:
        source_subp.end_reading()
        target_subp.end_writing()
        target_subp.cancel_reservation()

    # Make all the switchboard notifications outside the locked region.
    for n in notifications:
        switchboard.notify(*n)

    # Update BCs
    target_subp.autoenableBCs()

    target_subp.changed()
    switchboard.notify("redraw")
        


meshmenu.addItem(oofmenu.OOFMenuItem(
    'Copy_Field_State',
    callback=_copyFieldState,
    threadable=oofmenu.THREADABLE,
    params=[whoville.WhoParameter('source',ooflib.engine.mesh.meshes,
                                  tip=parameter.emptyTipString),
            whoville.WhoParameter('target',ooflib.engine.mesh.meshes,
                                  tip=parameter.emptyTipString)],
    help="Copy the Field state (defined, active, etc) from one Mesh to another.",
    discussion="""<para>

    This command copies the &field; state from the default
    &subproblem; in one &mesh; to another, meaning that the same
    &fields; will be defined, active, and in-plane in the
    <varname>target</varname> &mesh; as in the
    <varname>source</varname> &mesh;.  If &fields; were explicitly
    <link linkend='MenuItem:OOF.Mesh.Initialize'>initialized</link> in
    the source &mesh;, the initializers will be copied, but the
    command does <emphasis>not</emphasis> copy the &field; values.
    (This is because the source and target meshes might have quite
    different geometries.)</para>

    <para>DEPRECATED.  Use <xref
    linkend='MenuItem:OOF.Subproblem.Copy_Field_State'/> instead.
    
    </para>"""
                        ) )

    

# Likewise for equation state.  This is also deprecated. See
# subproblemmenu.py for the preferred version.

def _copyEquationState(menuitem, source, target):
    if source == target:
        raise ooferror.ErrUserError('Source and target must differ!')
    if parallel_enable.enabled():
        meshIPC.ipcmeshmenu.Copy_Equation_State(source=source,target=target)
        return

    notifications = []
    source_subp = ooflib.engine.mesh.meshes[source].get_default_subproblem()
    target_subp = ooflib.engine.mesh.meshes[target].get_default_subproblem()
    source_subp.begin_reading()
    target_subp.reserve()
    target_subp.begin_writing()
    try:
        source_obj = source_subp.getObject()
        target_obj = target_subp.getObject()
        source_eqns = source_obj.all_equations()
        target_eqns = target_obj.all_equations()

        for e in target_eqns:
            if not source_obj.is_active_equation(e):
                target_obj.deactivate_equation(e)
                notifications.append(
                    ("equation activated", target_subp.path(), e.name(), 0))
            if config.devel()>=1:
                if not source_obj.is_kinetically_active_equation(e):
                    target_obj.kinetic_deactivate_equation(e)
                    notifications.append(
                        ('kinetics activated', target_subp.path(), e.name(), 0))
                if not source_obj.is_dynamically_active_equation(e):
                    target_obj.deactivate_dynamics(e)
                    notifications.append(
                        ('dynamics activated', target_subp.path(), e.name(), 0))
        for e in source_eqns:
            if not target_obj.is_active_equation(e):
                target_obj.activate_equation(e)
                notifications.append(
                        ("equation activated", target_subp.path(), e.name(), 1))
            if config.devel()>=1:
                if not target_obj.is_kinetically_active_equation(e):
                    target_obj.kinetic_activate_equation(e)
                    notifications.append(
                        ('kinetics activated', target_subp.path(), e.name(), 1))
                if not target_obj.is_dynamically_active_equation(e):
                    target_obj.activate_dynamics(e)
                    notifications.append(
                        ('dynamics activated', target_subp.path(), e.name(), 1))
    finally:
        source_subp.end_reading()
        target_subp.end_writing()
        target_subp.cancel_reservation()

    for n in notifications:
        switchboard.notify(*n)

    target_subp.autoenableBCs()
    target_subp.changed()
    

meshmenu.addItem(oofmenu.OOFMenuItem(
    'Copy_Equation_State',
    callback=_copyEquationState,
    threadable=oofmenu.THREADABLE,
    params=[whoville.WhoParameter('source',ooflib.engine.mesh.meshes,
                                  tip=parameter.emptyTipString),
            whoville.WhoParameter('target',ooflib.engine.mesh.meshes,
                                  tip=parameter.emptyTipString)],
    help="Copy the set of active Equations from one Mesh to another.",
    discussion="""<para>

    This command copies the &equation; state from the default
    &subproblem; in one &mesh; to the default &subproblem; in another,
    meaning that the same &equations; will be active in the
    <varname>target</varname> &subproblem; as in the
    <varname>source</varname> &subproblem;.

    </para>"""
    
    ) )

# When using subproblems, the field initializers have to be copied
# separately from the Field state, because the initializers live
# in the mesh and the state flags live in the subproblems.

def _copyFieldInits(menuitem, source, target):
    if source == target:
        return
    if parallel_enable.enabled():
        meshIPT.ipcmeshmenu.Copy_Field_Initializers(source=source,target=target)
        return
    notifications=[]
    source_mesh = ooflib.engine.mesh.meshes[source]
    target_mesh = ooflib.engine.mesh.meshes[target]
    source_mesh.begin_reading()
    target_mesh.reserve()
    target_mesh.begin_writing()
    try:
        source_fields = source_mesh.all_subproblem_fields()
        target_fields = target_mesh.all_subproblem_fields()
        for f in source_fields:
            if f in target_fields:
                try:
                    initializer=source_mesh.initializers[f]
                except KeyError:
                    pass
                else:
                    target_mesh.initialize_field(f, initializer)
                    notifications.append(("field initialized"))
    finally:
        source_mesh.end_reading()
        target_mesh.end_writing()
        target_mesh.cancel_reservation()
    for n in notifications:
        switchboard.notify(*n)

meshmenu.addItem(oofmenu.OOFMenuItem(
    'Copy_Field_Initializers',
    callback=_copyFieldInits,
    threadable=oofmenu.THREADABLE,
    params=[whoville.WhoParameter('source', ooflib.engine.mesh.meshes,
                                  tip=parameter.emptyTipString),
            whoville.WhoParameter('target', ooflib.engine.mesh.meshes,
                                  tip=parameter.emptyTipString)],
    help="Copy all of the relevant Field initializers from one Mesh to another.",
    discussion="""<para>

    Copy all of the &field; initialization functions from the source
    &mesh; to the target &mesh;, and reinitialize the &fields;.  If a
    &field; is not defined in the target &mesh;, its initializer will
    not be copied.

    </para>"""
    ))

def _clearFieldInits(menuitem, mesh):
    themesh = ooflib.engine.mesh.meshes[mesh]
    themesh.reserve()
    themesh.begin_writing()
    try:
        for field in themesh.all_subproblem_fields():
            themesh.remove_initializer(field)
    finally:
        themesh.end_writing()
        themesh.cancel_reservation()
    switchboard.notify("field initialized")

meshmenu.addItem(oofmenu.OOFMenuItem(
    'Clear_Field_Initializers',
    callback=_clearFieldInits,
    params=[whoville.WhoParameter('mesh', ooflib.engine.mesh.meshes,
                                  tip=parameter.emptyTipString)],
    help="Remove all Field initializers from the current Mesh.",
    discussion="""<para>

    Remove all the &field; initializers from the given &mesh;.  This
    does not change the values of the &fields; themselves, but
    prevents them from being reinitialized later.

    </para>"""
    ))
        
    

#######################################

fieldmenu = meshmenu.addItem(oofmenu.OOFMenuItem(
    'Field',
    help='Define and activate Fields.',
    discussion="""<para>
    The <command>Field</command> menu contains the commands that
    define and set the properties of &fields; on &meshes;.
    </para>"""))


def _defineField(menuitem, mesh, field):
    ## This has been rewritten to use the default subproblem, for
    ## backwards compatibility.  The menuitem is deprecated -- use
    ## Subproblem.Field.Define instead.
    if parallel_enable.enabled():
        meshIPC.ipcfieldmenu.Define(mesh=mesh,field=field)
    else:
        meshcontext = ooflib.engine.mesh.meshes[mesh] 
        subpcontext = meshcontext.get_default_subproblem()
        subpcontext.reserve()
        subpcontext.begin_writing()
        try:
            subpcontext.getObject().define_field(field)
        finally:
            subpcontext.end_writing()
            subpcontext.cancel_reservation()

        switchboard.notify("field defined", subpcontext.path(), field.name(), 1)
        subpcontext.autoenableBCs()
        subpcontext.changed()
        
def _undefineField(menuitem, mesh, field):
    ## Also deprecated.  Use Subproblem.Field.Undefine instead.
    if parallel_enable.enabled():
        meshIPC.ipcfieldmenu.Undefine(mesh=mesh,field=field)
    else:
        meshcontext = ooflib.engine.mesh.meshes[mesh]
        subpcontext = meshcontext.get_default_subproblem()
        subpcontext.reserve()
        subpcontext.begin_writing()
        try:
            subpcontext.getObject().undefine_field(field)
        finally:
            subpcontext.end_writing()
            subpcontext.cancel_reservation()
            
        subpcontext.autoenableBCs()
        subpcontext.changed()
        switchboard.notify("field defined", subpcontext.path(), field.name(), 0)
    
fieldmenu.addItem(oofmenu.OOFMenuItem(
    'Define',
    threadable=oofmenu.THREADABLE,
    callback=_defineField,
    params=[whoville.WhoParameter('mesh', ooflib.engine.mesh.meshes,
                                  tip=parameter.emptyTipString),
            meshparameters.FieldParameter('field', tip=parameter.emptyTipString)
    ],
    help="Define a Field on a Mesh. Only defined Fields may be given values.",
    ## TODO: Fix discussion
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu/definefield.xml')
    ))

fieldmenu.addItem(oofmenu.OOFMenuItem(
    'Undefine',
    threadable=oofmenu.THREADABLE,
    callback=_undefineField,
    params=[whoville.WhoParameter('mesh', ooflib.engine.mesh.meshes,
                                  tip=parameter.emptyTipString),
            meshparameters.FieldParameter('field', tip=parameter.emptyTipString)
            ],
    help="Undefine a Field on a Mesh.  Only defined Fields may be given values.",
    discussion="""<para>

    Undefine a &field; on a &mesh;'s default &subproblem;.  This frees
    the memory used to store the &field; components and destroys their
    values, unless other &subproblems; are using the &field;.  See <xref
    linkend='MenuItem:OOF.Mesh.Field.Define'/>. DEPRECATED.

    </para>"""
    ))


def _activateField(menuitem, mesh, field):
    activation = False
    if parallel_enable.enabled():
        meshIPC.ipcfieldmenu.Activate(mesh=mesh,field=field)
    else:
        meshcontext = ooflib.engine.mesh.meshes[mesh]
        subpcontext = meshcontext.get_default_subproblem()
        subpcontext.reserve()
        subpcontext.begin_writing()
        try:
            subp = subpcontext.getObject()
            if subp.is_defined_field(field):
                subp.activate_field(field)
                activation = True
            else:
                reporter.report(
                    "You must define a Field before you can activate it.")
        finally:
            subpcontext.end_writing()
            subpcontext.cancel_reservation()
            
        if activation:
            subpcontext.autoenableBCs()
            switchboard.notify("field activated", subpcontext.path(),
                               field.name(), 1)
            subpcontext.changed()
    
def _deactivateField(menuitem, mesh, field):
    deactivation = False
    if parallel_enable.enabled():
        meshIPC.ipcfieldmenu.Deactivate(mesh=mesh,field=field)
    else:
        meshcontext = ooflib.engine.mesh.meshes[mesh]
        subpcontext = meshcontext.get_default_subproblem()
        subpcontext.reserve()
        subpcontext.begin_writing()
        try:
            subp = subpcontext.getObject()
            if subp.is_active_field(field):
                subp.deactivate_field(field)
                deactivation = True
            else:
                reporter.report(
                    "You must define and activate a Field before you can deactivate it.")
        finally:
            subpcontext.end_writing()
            subpcontext.cancel_reservation()
            
        if deactivation:
            subpcontext.autoenableBCs()
            switchboard.notify("field activated", subpcontext.path(),
                               field.name(), 0)
            subpcontext.changed()
    
fieldmenu.addItem(oofmenu.OOFMenuItem(
    "Activate",
    threadable=oofmenu.THREADABLE,
    callback=_activateField,
    params=[whoville.WhoParameter('mesh', ooflib.engine.mesh.meshes,
                                  tip=parameter.emptyTipString),
            meshparameters.FieldParameter('field', tip=parameter.emptyTipString)
    ],
    help="Activate a Field.  The solver finds the values of active Fields.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu/activatefield.xml')
    ))

fieldmenu.addItem(oofmenu.OOFMenuItem(
    'Deactivate',
    threadable=oofmenu.THREADABLE,
    callback=_deactivateField,
    params=[whoville.WhoParameter('mesh', ooflib.engine.mesh.meshes,
                                  tip=parameter.emptyTipString),
            meshparameters.FieldParameter('field', tip=parameter.emptyTipString)
            ],
    help="Deactivate a Field.  The solver finds the values of active Fields.",
    discussion="""<para>

    Deactivating a &field; means that its values will not be found
    when the &mesh; is <link
    linkend="MenuItem:OOF.Solver.Solve">solved</link>.  See <xref
    linkend='MenuItem:OOF.Mesh.Field.Activate'/>.

    </para>"""
    ))

def _inPlaneField(menuitem, mesh, field):
    if parallel_enable.enabled():
        meshIPC.ipcfieldmenu.In_Plane(mesh=mesh,field=field)
    else:
        meshcontext = ooflib.engine.mesh.meshes[mesh]
        meshcontext.reserve()
        meshcontext.begin_writing()
        try:
            meshcontext.set_in_plane_field(field, 1)
        finally:
            meshcontext.end_writing()
            meshcontext.cancel_reservation()
        switchboard.notify("field inplane", meshcontext.path(), field.name(), 1)
        meshcontext.changed()
    
def _outOfPlaneField(menuitem, mesh, field):
    if parallel_enable.enabled():
        meshIPC.ipcfieldmenu.Out_of_Plane(mesh=mesh,field=field)
    else:
        meshcontext = ooflib.engine.mesh.meshes[mesh]
        meshcontext.reserve()
        meshcontext.begin_writing()
        try:
            meshcontext.set_in_plane_field(field, 0)
        finally:
            meshcontext.end_writing()
            meshcontext.cancel_reservation()
        switchboard.notify("field inplane", meshcontext.path(),
                           field.name(), 0)    
        meshcontext.changed()
    
fieldmenu.addItem(oofmenu.OOFMenuItem(
    'In_Plane',
    threadable=oofmenu.THREADABLE,
    callback=_inPlaneField,
    params=[whoville.WhoParameter('mesh', ooflib.engine.mesh.meshes,
                                  tip=parameter.emptyTipString),
            meshparameters.FieldParameter('field', tip=parameter.emptyTipString)
            ],
    help="In-plane Fields are constrained to have no z-components.",
    discussion="""<para>

    This command invokes <link
    linkend='Section:Concepts:Mesh:3D'>generalized plane-strain</link>
    for the given &field; on all &subproblems; on the given &mesh;.
    The out-of-plane derivatives of the &field; are taken to be zero.
    See <xref linkend='MenuItem:OOF.Mesh.Field.Out_of_Plane'/>.>
    
    </para>"""
    ))

fieldmenu.addItem(oofmenu.OOFMenuItem(
    'Out_of_Plane',
    threadable=oofmenu.THREADABLE,
    callback=_outOfPlaneField,
    params=[whoville.WhoParameter('mesh', ooflib.engine.mesh.meshes,
                                  tip=parameter.emptyTipString),
            meshparameters.FieldParameter('field', tip=parameter.emptyTipString)
            ],
    help="Out-of-plane Fields are allowed to have z-components.",
    discussion="""<para>

    This command disables <link
    linkend='Section:Concepts:Mesh:3D'>generalized plane-strain</link>
    for the given &field; on all &subproblems; on the given &mesh;.
    The out-of-plane derivatives of the &field; will be computed.
    Generally, it's necessary to <link
    linkend='MenuItem:OOF.Mesh.Equation.Activate'>activate</link> a
    <link
    linkend='Section:Concepts:Mesh:Equation:PlaneFlux'>plane-flux
    equation</link> in order to solve for the out-of-plane derivatives
    of a &field;.
    
    </para>"""
    ))

####################

eqnmenu = meshmenu.addItem(oofmenu.OOFMenuItem('Equation',
                                               help='Activate equations.'))

def _activateEquation(menuitem, mesh, equation):
    if parallel_enable.enabled():
        meshIPC.ipceqnmenu.Activate(mesh=mesh,equation=equation)
    else:
        meshcontext = ooflib.engine.mesh.meshes[mesh]
        subpcontext = meshcontext.get_default_subproblem()
        subpcontext.reserve()
        subpcontext.begin_writing()
        try:
            subpcontext.getObject().activate_equation(equation)
        finally:
            subpcontext.end_writing()
            subpcontext.cancel_reservation()

        subpcontext.autoenableBCs()
        switchboard.notify('equation activated', subpcontext.path(),
                           equation.name(), 1)
        subpcontext.changed()
    
def _deactivateEquation(menuitem, mesh, equation):
    if parallel_enable.enabled():
        meshIPC.ipceqnmenu.Deactivate(mesh=mesh,equation=equation)
    else:
        meshcontext = ooflib.engine.mesh.meshes[mesh]
        subpcontext = meshcontext.get_default_subproblem()
        subpcontext.reserve()
        subpcontext.begin_writing()
        try:
            subpcontext.getObject().deactivate_equation(equation)
        finally:
            subpcontext.end_writing()
            subpcontext.cancel_reservation()

        switchboard.notify('equation activated', subpcontext.path(),
                           equation.name(), 0)
        subpcontext.autoenableBCs()
        subpcontext.changed()
    
eqnmenu.addItem(oofmenu.OOFMenuItem(
    'Activate',
    callback=_activateEquation,
    threadable=oofmenu.THREADABLE,
    params=[whoville.WhoParameter('mesh', ooflib.engine.mesh.meshes,
                                  tip=parameter.emptyTipString),
            meshparameters.EquationParameter('equation',
                                             tip=parameter.emptyTipString)
    ],
    help="Activate an Equation.  The Solver solves the active Equations.",
    discussion="""<para>

    Activate the given &equation; on the default &subproblem; on the
    given &mesh;. Activated &equations; are the ones that will be
    <link linkend='MenuItem:OOF.Solver.Solve'>solved</link>.  For a
    solution to be possible, the active &equations; must involve
    &fluxes; that are produced by &properties; in the &mesh;, and
    those &properties; must couple to <link
    linkend='MenuItem:OOF.Mesh.Field.Define'>defined</link> &fields;.
    There must be as many active &equations; as there are <link
    linkend='MenuItem:OOF.Mesh.Field.Activate'>active</link> &fields;</para>

    <para> DEPRECATED. Use <xref
    link="MenuItem:OOF.Subproblem.Equation.Activate"/> instead.

    </para>"""
    ))

eqnmenu.addItem(oofmenu.OOFMenuItem(
    'Deactivate',
    callback=_deactivateEquation,
    threadable=oofmenu.THREADABLE,
    params=[whoville.WhoParameter('mesh', ooflib.engine.mesh.meshes,
                                  tip=parameter.emptyTipString),
            meshparameters.EquationParameter('equation',
                                             tip=parameter.emptyTipString)
    ],
    help="Deactivate an Equation.  The Solver solves the active Equations.",
    discussion="""<para>

    Deactivate the given &equation; on the default &subproblem; on the
    given &mesh;.  See <xref
    linkend='MenuItem:OOF.Mesh.Equation.Deactivate'/>.</para>

    <para> DEPRECATED.  USE <xref
    link="MenuItem:OOF.Subproblem.Equation.Deactivate"/> instead.
    
    </para>"""

    ))

#######################################
## Kinetic Equation Activation and Deactivation
## STARTS HERE
#######################################

## TODO: Create subproblem versions of the kinetics and dynamics menu
## items.  Fix these so that they work with old data files. (Is that
## necessary?  There shouldn't be any old data files in the wild that
## use these menu items.)

def _kineticActivateEquation(menuitem, mesh, equation):
    meshcontext = ooflib.engine.mesh.meshes[mesh]
    meshcontext.reserve()
    meshcontext.begin_writing()
    try:
        subp = meshcontext.femesh().get_default_subproblem()
        subp.kinetic_activate_equation(equation)
    finally:
        meshcontext.end_writing()
        meshcontext.cancel_reservation()

    switchboard.notify('kinetics activated', mesh, equation.name(), 1)
    ooflib.engine.mesh.meshes[mesh].changed()
    
def _kineticDeactivateEquation(menuitem, mesh, equation):
    meshcontext = ooflib.engine.mesh.meshes[mesh]
    meshcontext.reserve()
    meshcontext.begin_writing()
    try:
        subp = meshcontext.femesh().get_default_subproblem()
        subp.kinetic_deactivate_equation(equation)
    finally:
        meshcontext.end_writing()
        meshcontext.cancel_reservation()

    switchboard.notify('kinetics activated', mesh, equation.name(), 0)
    ooflib.engine.mesh.meshes[mesh].changed()

if config.devel()>=1:
    eqnmenu.addItem(oofmenu.OOFMenuItem(
        'ActivateKinetics',
        callback=_kineticActivateEquation,
        threadable=oofmenu.THREADABLE,
        params=[whoville.WhoParameter('mesh', ooflib.engine.mesh.meshes,
                                      tip=parameter.emptyTipString),
                meshparameters.EquationParameter('equation',
                                                 tip=parameter.emptyTipString)
        ]
        ))

    eqnmenu.addItem(oofmenu.OOFMenuItem(
        'DeactivateKinetics',
        callback=_kineticDeactivateEquation,
        threadable=oofmenu.THREADABLE,
        params=[whoville.WhoParameter('mesh', ooflib.engine.mesh.meshes,
                                      tip=parameter.emptyTipString),
                meshparameters.EquationParameter('equation',
                                                 tip=parameter.emptyTipString)
        ]
        ))


#######################################
## Kinetic Equation Activation and Deactivation
## ENDS HERE
#######################################



#######################################
## Activation and Deactivation of Dynamic Term in Equation 
## STARTS HERE
#######################################

def _activateDynamics(menuitem, mesh, equation):
    meshcontext = ooflib.engine.mesh.meshes[mesh]
    meshcontext.reserve()
    meshcontext.begin_writing()
    try:
        subp = meshcontext.femesh().get_default_subproblem()
        subp.activate_dynamics(equation)
    finally:
        meshcontext.end_writing()
        meshcontext.cancel_reservation()
    switchboard.notify('dynamics activated', mesh, equation.name(), 1)
    ooflib.engine.mesh.meshes[mesh].changed()
    
def _deactivateDynamics(menuitem, mesh, equation):
    meshcontext = ooflib.engine.mesh.meshes[mesh]
    meshcontext.reserve()
    meshcontext.begin_writing()
    try:
        subp = meshcontext.femesh().get_default_subproblem()
        subp.deactivate_dynamics(equation)
    finally:
        meshcontext.end_writing()
        meshcontext.cancel_reservation()
    switchboard.notify('dynamics activated', mesh, equation.name(), 0)
    ooflib.engine.mesh.meshes[mesh].changed()

if config.devel()>=1:
    eqnmenu.addItem(oofmenu.OOFMenuItem(
        'ActivateDynamics',
        callback=_activateDynamics,
        threadable=oofmenu.THREADABLE,
        params=[whoville.WhoParameter('mesh', ooflib.engine.mesh.meshes,
                                      tip=parameter.emptyTipString),
                meshparameters.EquationParameter('equation',
                                                 tip=parameter.emptyTipString)
        ],
        help="Enable second order time derivatives."
        ))

    eqnmenu.addItem(oofmenu.OOFMenuItem(
        'DeactivateDynamics',
        callback=_deactivateDynamics,
        threadable=oofmenu.THREADABLE,
        params=[whoville.WhoParameter('mesh', ooflib.engine.mesh.meshes,
                                      tip=parameter.emptyTipString),
                meshparameters.EquationParameter('equation',
                                                 tip=parameter.emptyTipString)
        ],
        help="Disable second order time derivatives."
        ))


#######################################
## Kinetic Equation Activation and Deactivation
## ENDS HERE
#######################################



csmenu = meshmenu.addItem(oofmenu.OOFMenuItem(
    'Cross_Section',
    help="Create and manipulate Mesh cross sections for plotting.",
    discussion=xmlmenudump.loadFile("DISCUSSIONS/engine/menu/cross_section.xml")
    ))

def csnameresolver(param, name):
    if param.automatic():
        basename = 'cs'
    else:
        basename = name
    meshname = param.group['mesh'].value
    if meshname is not None:
        meshpath = labeltree.makePath(meshname)
        meshobj = ooflib.engine.mesh.meshes[meshpath]
        return meshobj.uniqueCSName(basename)

csnameparam = parameter.AutomaticNameParameter(
    'name', value=automatic.automatic, tip="Name of the cross section.",
    resolver=csnameresolver)

csmeshparam = whoville.WhoParameter('mesh', ooflib.engine.mesh.meshes,
                                    tip=parameter.emptyTipString)

csparamgroup = parameter.ParameterGroup(csnameparam, csmeshparam)

def _newCS(menuitem, mesh, name, cross_section):
    meshobj = ooflib.engine.mesh.meshes[mesh]
    meshobj.reserve()
    meshobj.begin_writing()
    try:
        meshobj.addCrossSection(name, cross_section)
    finally:
        meshobj.end_writing()
        meshobj.cancel_reservation()
    switchboard.notify("redraw")

csmenu.addItem(oofmenu.OOFMenuItem(
    'New',
    threadable=oofmenu.THREADABLE,
    callback=_newCS,
    params=csparamgroup + [
    parameter.RegisteredParameter('cross_section',
                                  meshcrosssection.MeshCrossSection,
                                  tip="New cross section object.") ],
    help="Create a new cross section on a Mesh.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu/new_cross_section.xml')
    ))

def _delCS(menuitem, mesh, name):
    meshobj = ooflib.engine.mesh.meshes[mesh]
    meshobj.reserve()
    meshobj.begin_writing()
    try:
        meshobj.removeCrossSection(name)
    finally:
        meshobj.end_writing()
        meshobj.cancel_reservation()
    switchboard.notify("redraw")
        
csmenu.addItem(oofmenu.OOFMenuItem(
    'Remove',
    threadable=oofmenu.THREADABLE,
    callback=_delCS,
    params=[whoville.WhoParameter('mesh', ooflib.engine.mesh.meshes,
                                  tip=parameter.emptyTipString),
            parameter.StringParameter('name', tip='Cross section to remove.')],
    help='Delete a cross section from a mesh.',
    discussion="""<para>
    Delete the cross section named <varname>name</varname> from the &mesh;
    named <varname>mesh</varname>.
    </para>"""))

def _selectCS(menuitem, mesh, cross_section):
    meshobj = ooflib.engine.mesh.meshes[mesh]
    meshobj.reserve()
    meshobj.begin_writing()
    try:
        meshobj.selectCrossSection(cross_section)
    finally:
        meshobj.end_writing()
        meshobj.cancel_reservation()

    switchboard.notify("redraw")

csmenu.addItem(oofmenu.OOFMenuItem(
    'Select',
    callback=_selectCS,
    params=[whoville.WhoParameter('mesh', ooflib.engine.mesh.meshes,
                                  tip=parameter.emptyTipString),
    parameter.StringParameter('cross_section', tip='Cross section to select.')],
    help="Select a cross section on a mesh.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu/select_cs.xml')
    ))

def _deselectCS(menuitem, mesh):
    meshobj = ooflib.engine.mesh.meshes[mesh]
    meshobj.reserve()
    meshobj.begin_writing()
    try:
        meshobj.deselectCrossSection()
    finally:
        meshobj.end_writing()
        meshobj.cancel_reservation()
    switchboard.notify("redraw")

csmenu.addItem(oofmenu.OOFMenuItem(
    'Deselect',
    callback=_deselectCS,
    params=[whoville.WhoParameter('mesh', ooflib.engine.mesh.meshes,
                                  tip=parameter.emptyTipString)],
    help="Deselect all cross sections on a mesh.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu/deselect_cs.xml')
    ))

def _renameCS(menuitem, mesh, cross_section, name):
    meshobj = ooflib.engine.mesh.meshes[mesh]
    meshobj.reserve()
    meshobj.begin_writing()
    try:
        meshobj.renameCrossSection(cross_section, name)
    finally:
        meshobj.end_writing()
        meshobj.cancel_reservation()

csmenu.addItem(oofmenu.OOFMenuItem(
    'Rename',
    callback=_renameCS,
    params=[whoville.WhoParameter('mesh', ooflib.engine.mesh.meshes,
                                  tip=parameter.emptyTipString),
            parameter.StringParameter('cross_section',
                                      tip='Cross section to rename.'),
            parameter.StringParameter('name',
                                      tip='New name for the cross section.')
            ],
    help="Rename a cross section on a mesh.",
    discussion="<para>Assign a new name to a cross section.</para>"))

def _editCS(menuitem, mesh, name, cross_section):
    meshobj = ooflib.engine.mesh.meshes[mesh]
    meshobj.reserve()
    meshobj.begin_writing()
    try:
        meshobj.replaceCrossSection(name, cross_section)
    finally:
        meshobj.end_writing()
        meshobj.cancel_reservation()
    switchboard.notify("redraw")

csmenu.addItem(oofmenu.OOFMenuItem(
    'Edit',
    callback=_editCS,
    params=[whoville.WhoParameter('mesh', ooflib.engine.mesh.meshes,
                                  tip=parameter.emptyTipString),
            parameter.StringParameter('name', tip='Cross section to edit.'),
            parameter.RegisteredParameter('cross_section',
                                          meshcrosssection.MeshCrossSection,
                                          tip='New value for the cross section.')
    ],
    help="Reparametrize a cross section on a mesh.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu/edit_cs.xml')
    ))

def _copyCS(menuitem, current, cross_section, mesh, name):
    sourcemesh = ooflib.engine.mesh.meshes[current]
    sourcemesh.begin_reading()
    try:
        cs = sourcemesh.getCrossSection(cross_section).clone()
    finally:
        sourcemesh.end_reading()
        
    targetmesh = ooflib.engine.mesh.meshes[mesh]
    targetmesh.reserve()
    targetmesh.begin_writing()
    try:
        targetmesh.addCrossSection(name,cs)
    finally:
        targetmesh.end_writing()
        targetmesh.cancel_reservation()
    

csmenu.addItem(oofmenu.OOFMenuItem(
    'Copy',
    callback=_copyCS,
    params=[whoville.WhoParameter('current', ooflib.engine.mesh.meshes,
                                  tip='Mesh to copy the cross section from.'),
            parameter.StringParameter('cross_section',
                                      tip='Cross section to copy.')
            ]
    + parameter.ParameterGroup(
    whoville.WhoParameter('mesh', ooflib.engine.mesh.meshes,
                          tip='Mesh to copy the cross section to.'),
    parameter.AutomaticNameParameter('name',
                                     value=automatic.automatic,
                                     resolver=csnameresolver,
                                     tip='Name of the copied cross section.')),
    help="Copy a cross section, possibly to a different Mesh.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu/copy_cs.xml')
    ))
    

#######################################

def initField(menuitem, mesh, field, initializer):
    # This routine is repeated almost verbatim in meshIO.py, where
    # it's used to initialize meshes loaded from files.
    meshcontext = ooflib.engine.mesh.meshes[mesh]
    meshcontext.reserve()
    meshcontext.begin_writing()
    try:
        meshcontext.initialize_field(field, initializer)
    finally:
        meshcontext.end_writing()
        meshcontext.cancel_reservation()
    switchboard.notify("field initialized")

    for subproblem in meshcontext.subproblems():
        if field in subproblem.all_fields():
            subproblem.changed()
    switchboard.notify("redraw")
    
meshmenu.addItem(oofmenu.OOFMenuItem(
    'Initialize',
    threadable=oofmenu.THREADABLE,
    callback = initField,
    params=[whoville.WhoParameter('mesh', ooflib.engine.mesh.meshes,
                                  tip=parameter.emptyTipString),
            meshparameters.FieldParameter('field',
                                          tip=parameter.emptyTipString,
                                          outofplane=1),
            fieldinit.FieldInitParameter('initializer',
                                         tip=parameter.emptyTipString)
    ],
    help="Assign values to a Field on a Mesh.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu/initfield.xml')
    ))


#####################

def saveMesh(menuitem, filename, mode, format, mesh):
    quit.disallow()
    try:
        meshcontext = ooflib.engine.mesh.meshes[mesh]
        meshcontext.begin_reading()
        try:
            meshpath = labeltree.makePath(mesh)
            skelpath = meshpath[:2]
            skelcontext = skeletoncontext.skeletonContexts[skelpath]
            if format==datafile.ABAQUS:
                meshIO.writeABAQUSfromMesh(filename, mode, meshcontext)
            else:
                dfile = datafile.writeDataFile(filename, mode, format)
                microstructureIO.writeMicrostructure(
                    dfile,
                    microstructure.getMSContextFromMS(skelcontext.getObject().MS))
                skeletonIO.writeSkeleton(dfile, skelcontext)
                meshIO.writeMesh(dfile, meshcontext, skelcontext)
                dfile.close()
        finally:
            meshcontext.end_reading()
    finally:
        quit.allow()

OOF.File.Save.addItem(oofmenu.OOFMenuItem(
    'Mesh',
    callback = saveMesh,
    ordering=80,
    threadable=oofmenu.THREADABLE,
    params = [
    parameter.StringParameter('filename', tip="Name of the file."),
    parameter.StringParameter('mode', tip="'w' to (over)write and 'a' to append."),
    enum.EnumParameter('format', datafile.DataFileFormatExt, datafile.ASCII,
                       tip="Format of the file."),
    whoville.WhoParameter('mesh', ooflib.engine.mesh.meshes,
                          tip='Name of the Mesh.')],
    help="Save a Mesh to a file.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu/savemesh.xml')
    ))

def _fixmenu(*args):
    if ooflib.engine.mesh.meshes.nActual() == 0:
        OOF.File.Save.Mesh.disable()
    else:
        OOF.File.Save.Mesh.enable()

_fixmenu()

switchboard.requestCallback(('new who', 'Mesh'), _fixmenu)
switchboard.requestCallback(('remove who', 'Mesh'), _fixmenu)

##########################

def modifyMesh(menuitem, mesh, modifier):
    # The structure is same as "skeletonmenu._modify()"
    ModifyProgressBar = menuitem.getProgressBar(
        modifier.get_progressbar_type())
    
    meshcontext = ooflib.engine.mesh.meshes[mesh]
    meshcontext.reserve()
    meshcontext.begin_writing()
    try:
        reporter.start_progressbar()
        modifier.apply(meshcontext)
        reporter.progressbar_completed()
        reporter.end_progressbar()
    finally:
        meshcontext.end_writing()
        meshcontext.cancel_reservation()
        if ModifyProgressBar.query_stop() or ModifyProgressBar.get_success()<0:
            ModifyProgressBar.set_failure()
            ModifyProgressBar.set_message("Failed.")
            return None
        else:
            ModifyProgressBar.set_success()
            ModifyProgressBar.set_message("Succeeded.")

    switchboard.notify("mesh changed", meshcontext)
    switchboard.notify('redraw')
    switchboard.notify('Mesh modified', mesh, modifier)

OOF.Mesh.addItem(oofmenu.OOFMenuItem(
    'Modify',
    callback=modifyMesh,
    threadable=oofmenu.THREADABLE,
    params=[
    whoville.WhoParameter('mesh', ooflib.engine.mesh.meshes, tip=parameter.emptyTipString),
    parameter.RegisteredParameter('modifier', meshmod.MeshModification,
                                  tip="Mesh modifier.")
    ],
    help="Make changes to a Mesh.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu/modify_mesh.xml')
    ))

#############################

# SC Patch Recovery

if config.devel()>=1:
    def recoverFluxes(menuitem, mesh):
        meshcontext = ooflib.engine.mesh.meshes[mesh]
        skel = meshcontext.getSkeleton()
        femesh = meshcontext.femesh()
        femesh.create_scpatch(skel)
        femesh.flux_recovery()
    
    OOF.Mesh.addItem(oofmenu.OOFMenuItem(
        'SCPRecovery',
        callback=recoverFluxes,
        threadable=oofmenu.THREADABLE,
        params=[
        whoville.WhoParameter('mesh', ooflib.engine.mesh.meshes,
                              tip=parameter.emptyTipString)],
        help="Superconvergent Patch Recovery."))
    
